#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of gluster-health-report project which is a
# subproject of GlusterFS ( www.gluster.org)
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import tempfile
import uuid
import json

from kubectl_gluster.utils import kubectl_get, template_kube_apply, \
    GlusterCSException, error, info, kubectl_exec, \
    kubectl_context_run, execute

# Manifest files
ManifestGcsNamespace = "gcs-namespace.yml"
ManifestEtcdOperator = "gcs-etcd-operator.yml"
ManifestEtcdCluster = "gcs-etcd-cluster.yml"
ManifestGd2Services = "gcs-gd2-services.yml"
ManifestGd2Node = "gcs-gd2.yml"
ManifestFsCSI = "gcs-fs-csi.yml"
ManifestVirtBlockCSI = "gcs-virtblock-csi.yml"
ManifestPromethesOperator = "gcs-prometheus-operator.yml"
ManifestPrometheusBundle = "gcs-prometheus-bundle.yml"
ManifestPrometheusKubeStateMetrics = "gcs-prometheus-kube-state-metrics.yml"
ManifestPrometheusKubeMetrics = "gcs-prometheus-kube-metrics.yml"
ManifestPrometheusEtcdMetrics = "gcs-prometheus-etcd.yml"
ManifestPrometheusNodeMetrics = "gcs-prometheus-node-exporter.yml"
ManifestPrometheusOperatorMetrics = "gcs-prometheus-operator-metrics.yml"
ManifestGrafanaMixins = "gcs-mixins.yml"
ManifestGrafanaDashboard = "gcs-grafana.yml"
ManifestPrometheusAlertManager = "gcs-prometheus-alertmanager-cluster.yml"
ManifestVirtBlockCSIStorageClass = "gcs-storage-virtblock.yml"
ManifestStorageSnapshot = "gcs-storage-snapshot.yml"


def gcs_namespace_setup(config):
    template_kube_apply(config, ManifestGcsNamespace)
    info("GCS Namespace created")


def etcd_setup(config):
    template_kube_apply(config, ManifestEtcdOperator)
    info("Etcd Operator created")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.availableReplicas",
        gettype="deployment",
        name="etcd-operator",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking etcd operator status"
    )

    info("Etcd Operator is ready")

    template_kube_apply(
        config,
        ManifestEtcdCluster,
        retries=5,
        delay=5,
        label="Checking Etcd Operator status"
    )
    info("Etcd cluster created")

    etcd_client_endpoint = kubectl_get(
        namespace=config["namespace"],
        jsonpath="spec.clusterIP",
        gettype="service",
        name="etcd-client",
        retries=5,
        delay=5,
        label="Fetch etcd Cluster IP"
    )

    def check_num_etcd_running(out):
        # 3 etcd pods + one header line in CLI
        return len(out.split("\n")) == 4

    etcd_client_endpoint = "http://%s:2379" % etcd_client_endpoint
    config["template-args"]["etcd_client_endpoint"] = etcd_client_endpoint
    info("Etcd Service IP is available", endpoint=etcd_client_endpoint)
    kubectl_get(
        namespace=config["namespace"],
        jsonpath="",
        gettype="pods",
        name="",
        extra_args=["-l", "etcd_cluster=etcd",
                    "--field-selector=status.phase=Running"],
        out_expect_fn=check_num_etcd_running,
        retries=50,
        delay=10,
        label="Check for Etcd cluster status"
    )

    info("Etcd pods are UP")

    def check_etcd_cluster_ready(cmdout):
        members_data = json.loads(cmdout.strip())
        return len(members_data.get("members", [])) == 3

    # Etcd port-forward so that status can be checked
    with kubectl_context_run(
            ["port-forward",
             "svc/etcd-client",
             "31000:2379",
             "-n%s" % config["namespace"]]):

        execute(
            ["curl", "http://localhost:31000/v2/members"],
            out_expect_fn=check_etcd_cluster_ready,
            retries=50,
            delay=10,
            label="Checking etcd cluster ready"
        )

    info("Etcd cluster is ready")


def glusterd2_setup(config):
    template_kube_apply(config, ManifestGd2Services)
    info("Glusterd2 services created")

    for node in config["nodes"]:
        config["template-args"]["kube_hostname"] = node["address"]
        node_manifest_file = ManifestGd2Node.replace(
            ".yml", "-" + node["address"] + ".yml")
        template_kube_apply(
            config,
            node_manifest_file,
            template_file=ManifestGd2Node
        )
        info("Glusterd2 pod created", address=node["address"])

    def check_num_glusterd2_running(out):
        # glusterd2 pods + one header line in CLI
        return len(out.split("\n")) == (len(config["nodes"]) + 1)

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="",
        gettype="pods",
        name="",
        extra_args=["-l", "app.kubernetes.io/name=glusterd2",
                    "--field-selector=status.phase=Running"],
        out_expect_fn=check_num_glusterd2_running,
        retries=50,
        delay=10,
        label="Check for Glusterd2 pods status"
    )
    info("glusterd2 pods are UP")

    gd2_client_endpoint = "http://gluster-%s-0:24007" % (
        config["nodes"][0]["address"]
    )

    def check_num_peers(cmdout):
        peers = json.loads(cmdout.strip())
        return len(peers) == len(config["nodes"])

    curl_cmd = "curl %s/v1/peers" % gd2_client_endpoint
    peers = kubectl_exec(
        config["namespace"],
        "gluster-%s-0" % config["nodes"][0]["address"],
        curl_cmd,
        out_expect_fn=check_num_peers,
        retries=50,
        delay=10,
        label="Fetching the Gluster Peer info"
    )

    info("Glusterd2 cluster is ready")

    peers_json = json.loads(peers)

    for peer in peers_json:
        kube_hostname = peer["name"].split('-')[1]
        config["template-args"]["kube_hostname"] = kube_hostname
        devices = []
        for node in config["nodes"]:
            if node["address"] == kube_hostname:
                devices = node["devices"]
                break

        for device in devices:
            cmd = "glustercli device add %s %s" % (peer["id"], device)
            kubectl_exec(
                config["namespace"],
                "gluster-%s-0" % config["nodes"][0]["address"],
                cmd,
                retries=5,
                delay=5,
                label="Adding device to peer"
            )
            info("Added device", peer=kube_hostname, device=device)


def fs_csi_setup(config):
    template_kube_apply(config, ManifestFsCSI)
    info("GlusterCS FS CSI driver pods created")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.readyReplicas",
        gettype="statefulset",
        name="csi-glusterfsplugin-provisioner",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking FS CSI Provisioner state"
    )
    info("FS CSI provisioner is Ready")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.readyReplicas",
        gettype="statefulset",
        name="csi-glusterfsplugin-attacher",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking FS CSI attacher status"
    )
    info("FS CSI attacher is Ready")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.numberAvailable",
        gettype="daemonset",
        name="csi-glusterfsplugin-nodeplugin",
        retries=50,
        delay=10,
        out_expect="%s" % config["cluster-size"],
        label="Checking CSI node plugins"
    )
    info("FS CSI node plugins are Ready")

    template_kube_apply(config, ManifestStorageSnapshot)
    info("FS Storage and Snapshot class created")


def virtblock_csi_setup(config):
    template_kube_apply(config, ManifestVirtBlockCSI)
    info("GlusterCS Virtblock CSI driver pods created")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.readyReplicas",
        gettype="statefulset",
        name="csi-glustervirtblock-provisioner",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking Virtblock CSI Provisioner state"
    )
    info("CSI Virtblock provisioner is Ready")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.readyReplicas",
        gettype="statefulset",
        name="csi-glustervirtblock-attacher",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking Virtblock CSI attacher status"
    )
    info("CSI Virtblock attacher is Ready")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.numberAvailable",
        gettype="daemonset",
        name="csi-glustervirtblock-nodeplugin",
        retries=50,
        delay=10,
        out_expect="%s" % config["cluster-size"],
        label="Checking Virtblock CSI node plugins"
    )
    info("Virtblock CSI node plugins are Ready")

    template_kube_apply(config, ManifestVirtBlockCSIStorageClass)
    info("Virtblock Storage class created")


def monitoring_setup(config):
    template_kube_apply(config, ManifestPromethesOperator)
    info("Prometheus operator created")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="status.availableReplicas",
        gettype="deployment",
        name="prometheus-operator",
        retries=50,
        delay=10,
        out_expect="1",
        label="Checking Prometheus Operator"
    )
    info("Prometheus operator is Ready")

    kubectl_get(
        namespace="",
        jsonpath="",
        gettype="customresourcedefinitions",
        name="servicemonitors.monitoring.coreos.com",
        retries=30,
        delay=10,
        label="Checking Prometheus CRD"
    )
    info("Prometheus CRD check successful")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="",
        gettype="servicemonitors",
        name="",
        retries=30,
        delay=10
    )
    info("Prometheus service monitor CRD check successful")

    kubectl_get(
        namespace=config["namespace"],
        jsonpath="",
        gettype="Prometheus",
        name="",
        retries=30,
        delay=10
    )
    info("Prometheus namespace check successful")

    monitoring_extra = (
        (
            "Prometheus services, ServiceMonitor and Prometheus Objects",
            ManifestPrometheusBundle
        ),
        (
            "Kube-State-Metrics Exporter",
            ManifestPrometheusKubeStateMetrics
        ),
        (
            "Kubelet, APIServer and CoreDNS",
            ManifestPrometheusKubeMetrics
        ),
        (
            "Etcd Operator and the etcd cluster Exporters",
            ManifestPrometheusEtcdMetrics
        ),
        (
            "Node exporter",
            ManifestPrometheusNodeMetrics
        ),
        (
            "Prometheus Operator Exporter",
            ManifestPrometheusOperatorMetrics
        ),
        (
            "Alert manager Cluster",
            ManifestPrometheusAlertManager
        ),
        (
            "Grafana mixins",
            ManifestGrafanaMixins
        ),
        (
            "Grafana dashboard",
            ManifestGrafanaDashboard
        )
    )

    for msg, filename in monitoring_extra:
        template_kube_apply(config, filename, retries=2, delay=10,
                            label="Checking status of %s" % msg)
        info(msg + " deployed")


def deploy(config):
    config["template-args"] = {}
    config["template-args"]["gcs_namespace"] = config["namespace"]

    config["template-args"]["gcs_gd2_clusterid"] = str(uuid.uuid1())

    if not config.get("manifests-dir", ""):
        config["manifests-dir"] = (
            "https://raw.githubusercontent.com/gluster/"
            "gcs/master/deploy/templates/gcs-manifests/"
        )

    # Create working directory
    config["workdir"] = tempfile.mkdtemp(prefix="glustercs-")

    try:
        # GCS namespace create
        gcs_namespace_setup(config)

        # Etcd Operator and Etcd Cluster create
        etcd_setup(config)

        # Glusterd2 and devices setup
        glusterd2_setup(config)

        # FS CSI drivers setup
        fs_csi_setup(config)

        # Virtblock CSI drivers setup
        virtblock_csi_setup(config)

        # Monitoring setup
        monitoring_setup(config)

    except GlusterCSException as err:
        print()
        error(err)



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of gluster-health-report project which is a
# subproject of GlusterFS ( www.gluster.org)
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

import os
import sys
import subprocess
import time
from datetime import datetime
from contextlib import contextmanager

import requests
from jinja2 import Template


kubectl_cmd = "kubectl"


def _msg(prefix, msg, **kwargs):
    message = "[%7s] %s " % (prefix,
                             datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
    message += "%s" % msg
    for k, v in kwargs.items():
        if v:
            message += "  %s=%s" % (k, v)

    return message


def info(msg, **kwargs):
    msg = _msg("INFO", msg, **kwargs)
    if kwargs.get("newline", True):
        msg += "\n"

    sys.stdout.write(msg)
    sys.stdout.flush()


def warn(msg, **kwargs):
    print(_msg("WARNING", msg, **kwargs))


def error(msg, **kwargs):
    print(_msg("ERROR", msg, **kwargs))
    sys.exit(1)


class GlusterCSException(Exception):
    def __init__(self, message, **kwargs):
        self.kwargs = kwargs
        self.message = message

        if sys.version_info >= (3,):
            super().__init__(message)
        else:
            super(GlusterCSException, self).__init__(message)

    def __str__(self):
        error(self.message, **self.kwargs)


class CmdException(GlusterCSException):
    pass


class TemplateException(GlusterCSException):
    pass


def execute(cmd, retries=0, delay=2, out_expect="", out_expect_fn=None,
            shell=False, label=""):
    p = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         shell=shell)
    out, err = p.communicate()
    if sys.version_info >= (3,):
        out = out.decode("utf-8").strip()
        err = err.decode("utf-8").strip()
    else:
        out = out.strip()
        err = err.strip()

    cmd_print = cmd if shell else " ".join(cmd)

    if p.returncode == 0:
        if out_expect and out.strip() == out_expect:
            return out

        if out_expect_fn is not None and out_expect_fn(out):
            return out

        if not out_expect and out_expect_fn is None:
            return out

    if retries > 0:
        sys.stdout.write(".")
        sys.stdout.flush()
        time.sleep(delay)
        return execute(cmd, retries=(retries-1), delay=delay,
                       out_expect=out_expect, out_expect_fn=out_expect_fn,
                       shell=shell, label=label)

    if p.returncode != 0:
        raise CmdException(
            "command failed",
            cmd=cmd_print,
            return_code=p.returncode,
            error=err
        )

    if out_expect and out_expect != out:
        raise CmdException("command output didn't match",
                           cmd=cmd_print,
                           return_code=p.returncode,
                           expected=out_expect,
                           actual=out)

    if out_expect_fn is not None and not out_expect_fn(out):
        raise CmdException("command output didn't match",
                           cmd=cmd_print,
                           return_code=p.returncode,
                           actual=out)

    return out


def kubectl_exec(namespace, podname, command, retries=0, delay=2,
                 out_expect="", out_expect_fn=None, label=""):
    if retries > 0 and label:
        info(label, newline=False)

    cmd = (kubectl_cmd + " exec -it " + podname + " /bin/bash -n" +
           namespace + " -- -c '%s'" % command)
    out = execute(
        cmd,
        retries=retries,
        delay=delay,
        out_expect=out_expect,
        out_expect_fn=out_expect_fn,
        shell=True,
        label=label
    )

    if retries > 0 and label:
        print(".")

    return out


def kubectl_create(config, filename, retries=0, delay=2, label=""):
    if retries > 0 and label:
        info(label, newline=False)

    out = execute([
        kubectl_cmd,
        "create",
        "-f",
        config["workdir"] + "/" + filename
    ], retries=retries, delay=delay, label=label)

    if retries > 0 and label:
        print(".")

    return out


def kubectl_get(namespace, jsonpath, gettype, name, retries=0, delay=2,
                out_expect="", out_expect_fn=None, extra_args=[], label=""):
    if retries > 0 and label:
        info(label, newline=False)

    cmd = [kubectl_cmd]
    if namespace:
        cmd += ["-n", namespace]

    if jsonpath:
        cmd += ["-ojsonpath={.%s}" % jsonpath]

    cmd += ["get", gettype]

    if name:
        cmd += [name]

    if extra_args:
        cmd += extra_args

    out = execute(cmd, retries=retries, delay=delay, out_expect=out_expect,
                  out_expect_fn=out_expect_fn, label=label)

    if retries > 0 and label:
        print(".")

    return out


def template(config, filename, template_file=None):
    tmpl_url = config["manifests-dir"] + "/" + filename + ".j2"
    dest = os.path.join(config["workdir"], filename)

    if template_file is not None:
        tmpl_url = config["manifests-dir"] + "/" + template_file + ".j2"

    content = ""
    if tmpl_url.startswith("http"):
        resp = requests.get(tmpl_url)
        if resp.status_code != 200:
            raise TemplateException("failed to fetch template",
                                    template=tmpl_url,
                                    status_code=resp.status_code)

        if sys.version_info >= (3,):
            content = resp.content.decode('utf-8')
        else:
            content = resp.content
    else:
        try:
            with open(tmpl_url) as f:
                content = f.read()
        except IOError as err:
            raise TemplateException("failed to open template file",
                                    template=tmpl_url,
                                    error=err)

    if not content:
        raise TemplateException("failed to generate template",
                                template=tmpl_url)

    try:
        Template(content).stream(
            **config["template-args"]).dump(dest)
        return
    except IOError as err:
        raise TemplateException("failed to save template output",
                                template=tmpl_url,
                                error=err)


def template_kube_apply(config, filename, retries=0, delay=2,
                        template_file=None, label=""):
    template(config, filename, template_file=template_file)
    kubectl_create(config, filename, retries=retries, delay=delay)


@contextmanager
def kubectl_context_run(cmd):
    cmd = [kubectl_cmd] + cmd
    p = subprocess.Popen(cmd, stderr=subprocess.PIPE,
                         stdout=subprocess.PIPE)
    try:
        # Give some time to run the given command
        time.sleep(5)
        yield p
    finally:
        p.terminate()



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of gluster-health-report project which is a
# subproject of GlusterFS ( www.gluster.org)
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

from argparse import ArgumentParser

from kubectl_gluster.utils import info, warn, error
from kubectl_gluster.deploy import deploy

import yaml


defaultClusterSize = 3
defaultNamespace = "gcs"


def get_args():
    parser = ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(dest="mode")

    # deploy
    parser_deploy = subparsers.add_parser('deploy')
    parser_deploy.add_argument("config", help="Config File name")
    return parser.parse_args()


def deploy_args_validate(config):
    # Hard coded to 3 for now, also added 1 to support minikube
    if config.get("cluster-size", 0) not in [defaultClusterSize, 1]:
        error("Invalid cluster-size.",
              actual=config.get("cluster-size", 0),
              required=defaultClusterSize)

    if not config.get("namespace", ""):
        config["namespace"] = defaultNamespace
        warn("Namespace not specified, using default namespace.",
             name=defaultNamespace)

    if len(config.get("nodes", [])) != config.get("cluster-size", 0):
        error("Invalid number of nodes provided",
              actual=len(config.get("nodes", [])),
              required=config.get("cluster-size", 0))

    for idx, node in enumerate(config["nodes"]):
        if not node.get("address", ""):
            error("Invalid address provided for one or more node")

        if len(node.get("devices", [])) == 0:
            warn("No raw devices provided", address=node["address"])

        info("Node details", address=node["address"],
             devices=",".join(node.get("devices", [])))

    info("Cluster Size", clusterSize=config["cluster-size"])
    info("Namespace", namespace=config["namespace"])


def do_deploy(args):
    try:
        config = yaml.load(open(args.config))
    except FileNotFoundError as err:
        error("Unable to open config file.", error=err)

    if isinstance(config, str):
        error("Unable to parse config file.")

    deploy_args_validate(config)
    deploy(config)


def main():
    try:
        args = get_args()
        if args.mode == "deploy":
            do_deploy(args)
    except KeyboardInterrupt:
        return


if __name__ == "__main__":
    main()



#
# Copyright (c) 2019 Red Hat, Inc.
#
# This file is part of gluster-health-report project which is a
# subproject of GlusterFS ( www.gluster.org)
#
# This file is licensed to you under your choice of the GNU Lesser
# General Public License, version 3 or any later version (LGPLv3 or
# later), or the GNU General Public License, version 2 (GPLv2), in all
# cases as published by the Free Software Foundation.

from setuptools import setup


setup(
    name="kubectl-gluster",
    version="0.9.4",
    packages=["kubectl_gluster"],
    include_package_data=True,
    install_requires=["jinja2", "PyYAML", "requests"],
    entry_points={
        "console_scripts": [
            "kubectl-gluster = kubectl_gluster.main:main"
        ]
    },
    platforms="linux",
    zip_safe=False,
    author="Gluster Developers",
    author_email="gluster-devel@gluster.org",
    description="GlusterCS deployment tool",
    license="GPLv2",
    keywords="gluster, container, kubernetes, glustercs",
    url="https://github.com/aravindavk/kubectl-gluster",
    long_description="""
    GlusterCS deployment tool
    """,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "Environment :: Console",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
    ],
)
